
import numpy as np
from torchvision import transforms
import torch

transformations = transforms.Compose([transforms.ToPILImage(),
                                      transforms.Resize(224),
                                      transforms.CenterCrop(224),
                                      transforms.ToTensor(),
                                      transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])



def prep_input_numpy(img:np.ndarray, device:str):
    """Preparing a Numpy Array as input to the NITEC classifier."""

    if len(img.shape) == 4:
        imgs = []
        for im in img:
            imgs.append(transformations(im))
        img = torch.stack(imgs)
    else:
        img = transformations(img)

    img = img.to(device)

    if len(img.shape) == 3:
        img = img.unsqueeze(0)

    return img



import torch
import torch.nn as nn
import torch.nn.functional as F
import math

class ResNet(nn.Module):
    def __init__(self, block, layers, fc_layers=2):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7)
        self.classifier = nn.Linear(512*block.expansion,fc_layers)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)

        x = self.classifier(x)        

        return x


import pathlib
from typing import Union
import torchvision

import cv2
import numpy as np
import torch
import torch.nn as nn
from face_detection import RetinaFace
import torch.nn.functional as F
import os 

from .results import NITECResultContainer
from .model import ResNet
from .utils import prep_input_numpy


class NITEC_Classifier:

    def __init__(
        self, 
        weights: pathlib.Path, 
        device: str = 'cpu', 
        include_detector:bool = True,
        confidence_threshold:float = 0.5
        ):

        # Save input parameters
        self.weights = weights
        self.include_detector = include_detector
        self.device = device
        self.confidence_threshold = confidence_threshold

        # Create NITEC model
        self.model = ResNet(torchvision.models.resnet.BasicBlock, [2, 2, 2, 2], 2)
        saved_state_dict = torch.load(os.path.join(self.weights), map_location=device)
        if 'model_state_dict' in saved_state_dict:
            self.model.load_state_dict(saved_state_dict['model_state_dict'])
        else:
            self.model.load_state_dict(saved_state_dict)
        self.model.to(self.device)
        self.model.eval()

        # Create RetinaFace if requested
        if self.include_detector:

            if device.type == 'cpu':
                self.detector = RetinaFace()
            else:
                self.detector = RetinaFace(gpu_id=device.index)

        
   
    def predict(self, frame: np.ndarray) -> NITECResultContainer:

        # Creating containers
        face_imgs = []
        bboxes = []
        landmarks = []
        scores = []
        results = np.empty((0,1))
        bboxes_stack = np.empty((0,1))
        landmarks_stack = np.empty((0,1))
        scores_stack = np.empty((0,1))

        if self.include_detector:
            faces = self.detector(frame)
            if faces is not None: 
                for box, landmark, score in faces:

                    # Apply threshold
                    if score < .7:
                        continue
                               
                    # Extract safe min and max of x,y
                    x_min=int(box[0])
                    if x_min < 0:
                        x_min = 0
                    y_min=int(box[1])
                    if y_min < 0:
                        y_min = 0
                    x_max=int(box[2])
                    y_max=int(box[3])
                    
                    # Crop image
                    img = frame[y_min:y_max, x_min:x_max]
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                    img = cv2.resize(img, (224, 224))
                    face_imgs.append(img)

                    bboxes.append(box)
                    landmarks.append(landmark)
                    scores.append(score)
            
                # Classify eye contact
                if len(face_imgs):
                    results = self.process_frame(np.stack(face_imgs))
                    bboxes_stack = np.stack(bboxes)
                    landmarks_stack = np.stack(landmarks)
                    scores_stack = np.stack(scores)
            else:

                results = np.empty((0,1))

        else:
            results = self.process_frame(frame)

        # Save data
        results = NITECResultContainer(
            results=results,
            bboxes=bboxes_stack,
            landmarks=landmarks_stack,
            scores=scores_stack
        )

        return results

    def process_frame(self, frame: Union[np.ndarray, torch.Tensor]):
        
        # Prepare input
        if isinstance(frame, np.ndarray):
            img = prep_input_numpy(frame, self.device)
        elif isinstance(frame, torch.Tensor):
            img = frame
        else:
            raise RuntimeError("Invalid dtype for input")
    
        # Classify
        output = self.model(img)
        output = F.sigmoid(output)
        val = output.cpu().numpy()[:,1]
        return val




from .results import NITECResultContainer
import numpy as np
import cv2


def visualize(frame: np.ndarray, results: NITECResultContainer, confidence=0.5):
    for i in range(results.results.shape[0]):

        bbox = results.bboxes[i]
        res = results.results[i]
        
        # Extract safe min and max of x,y
        x_min=int(bbox[0])
        if x_min < 0:
            x_min = 0
        y_min=int(bbox[1])
        if y_min < 0:
            y_min = 0
        x_max=int(bbox[2])
        y_max=int(bbox[3])

        circle_center = (int(x_min +(x_max-x_min)/2.0), int(y_min +(y_max-y_min)/2))
        circle_size= int((x_max-x_min)/2.0)
        vis_frame = frame.copy()

        if res >= confidence:                    
        # cv2.rectangle(frame, (x_min, y_min) , (x_max, y_max), (0,255,0), 5)
            cv2.circle(vis_frame, circle_center, circle_size, (0,255,0),-1)
        else:
            cv2.circle(vis_frame, circle_center, circle_size,(0,0,255),-1)

            #cv2.rectangle(frame, (x_min, y_min) , (x_max, y_max), (0,0,255), 5)
        frame = cv2.addWeighted(frame,0.7,vis_frame,0.3,0)

    return frame

        
